#!/usr/bin/env python3

from bs4 import BeautifulSoup
import re
from . import tlsutil, gaussian94
import json


""" Dictionary of the basis set formats supported by emsl as well as this script,
    mapped to the default file extension used.
"""
formats = {
    "Gaussian94": "g94",
}


class EmslError(Exception):
    """
    Generic error thrown if some data obtained from the EMSL basis set exchange
    is not in the format expected
    """
    def __init__(self, message):
        super(EmslError, self).__init__(message)


"""Cache for the base url"""
__base_url_cache = None


def get_base_url():
    """
    Unfortunately the emsl base url changes from time to time.
    This function determines it.
    """
    global __base_url_cache
    if __base_url_cache is not None:
        return __base_url_cache

    portal_url = "https://bse.pnl.gov/bse/portal"
    ret = tlsutil.get_tls_fallback(portal_url)

    if not ret.ok:
        raise EmslError("Error determining base url from {}.".format(portal_url))
    soup = BeautifulSoup(ret.text, "lxml")

    iframe = soup.find("iframe", class_="chefContentIFrame")
    if iframe is None:
        raise EmslError("Could not find content iframe in {}.".format(portal_url))

    title_url = iframe["src"]
    if not title_url.endswith("/panel/Main/template/content"):
        raise EmslError("Unexpected title iframe url")

    __base_url_cache = title_url[:-28]
    return __base_url_cache


def _parse_list(string):
    """ Parse a list string into a list of strings"""
    if string[0] != "[" and string[-1] != "]":
        raise ValueError("Invalid list string: " + string)
    return [elem.strip() for elem in string[1:-1].split(sep=",")]


def _parse_basis_line(line):
    startstr = "new basisSet("
    endstr = ");"

    # Truncate the line
    line = line[line.find(startstr) + len(startstr):line.rfind(endstr)]

    # And split into argument strings (without leading and tailling '"')
    splitted = [m[1:-1] for m in re.findall('"[^"]*"', line)]

    if len(splitted) != 11:
        raise ValueError("Invalid emsl basis line: " + line +
                         ", error: More params than expected.")

    try:
        elems = _parse_list(splitted[3])
    except ValueError as e:
        raise ValueError("Invalid emsl basis line: " + line + ", error: " + e.args[0])

    return {
        "url": splitted[0],              # Url to download it from
        "name": splitted[1],             # Name of the basis
        "type": splitted[2],
        "elements": elems,
        "status": splitted[4],
        "hasEcp": splitted[5],
        "hasSpin": splitted[6],
        "lastModifiedDate": splitted[7],
        "contributionPI": splitted[8],
        "contributorName": splitted[9],
        "description": splitted[10]      # Short description of the basis
    }


def download_basisset_list(return_elements=False):
    """
    Download and parse the list of basis sets from emsl
    Returns a list of basis set dictionaries
    """
    # TODO This function should go in the future and merged
    #      into add_to_database

    base_url = get_base_url()
    ret = tlsutil.get_tls_fallback(base_url + "/panel/Main/template/content")

    if not ret.ok:
        raise EmslError("Error downloading list of basis sets from emsl")
    soup = BeautifulSoup(ret.text, "lxml")

    basis_sets = []  # The basis set list to return

    # Search expression for script tags which define basisSet objects
    re_bassets = re.compile("basisSets\[[0-9]+\]\W*=")

    # Search expression for the basisSet definition lines:
    re_basdef = re.compile("^\W*basisSets\[[0-9]+\]\W*=\W*new\W*basisSet")

    # Search expression for the number of basis sets expected
    re_num = re.compile("numBasis\W*=\W*([0-9]+)")

    # Seek through all script blocks, which contain basis definitions:
    for script in soup.find_all("script"):
        # Ignore script html tags, which do not contain the string
        # 'basisSets[number]=' in their text
        if not re_bassets.search(script.text):
            continue
        lines = script.text.splitlines()

        numlines = [re_num.search(l).group(1) for l in lines if re_num.search(l)]
        if len(numlines) > 1:
            raise EmslError("The string describing the number of basis sets is "
                            "found more than once.")
        expected_num_bases = int(numlines[0])

        try:
            bases = [_parse_basis_line(l) for l in lines if re_basdef.match(l)]
        except ValueError as e:
            raise EmslError(e.args[0])

        if (len(bases) != expected_num_bases):
            raise EmslError("Deviation between expected number of basis definitions "
                            "and the actual number found.")
        basis_sets.extend(bases)

    if len(basis_sets) == 0:
        raise EmslError("No basis sets obtained from emsl bse data")

    if not return_elements:
        return basis_sets

    elements = []  # The element list to return
    for div in soup.find_all(class_="table-row", name="div"):
        for elem in div.find_all(class_="elt", name="a"):
            # Loop over the periodic table that the emsl start page
            # produces and extract the elements

            if "id" not in elem.attrs or "title" not in elem.attrs:
                raise EmslError("Elements of the periodic table does not "
                                "contain attribute "
                                "'id' or 'title': {}".format(str(elem)))
            try:
                atnum = int(elem["id"])
            except ValueError:
                raise EmslError("Elements of the periodic table have a non-integer id "
                                "which cannot be interpreted as the atomic number: "
                                "{}".format(str(elem)))

            elements.append({
                "symbol": elem.text,
                "atnum": atnum,
                "name": elem["title"]
            })
    if len(elements) == 0:
        raise EmslError("No elements obtained from emsl bse data")
    return elements, basis_sets


def add_to_database(db):
    """
    Add the basis set definitions to the database
    """
    elements, lst = download_basisset_list(return_elements=True)
    db.create_table_of_elements("EMSL", elements)

    for bas in lst:
        extra = json.dumps({"url": bas["url"]})
        basset_id = db.insert_basisset(bas["name"], source="EMSL", extra=extra,
                                       description=bas["description"])

        for atom in bas["elements"]:
            try:
                element = db.search_element("EMSL", atom)
            except ValueError as e:
                if atom != "X":
                    # atom == X is a known issue in some of the basis set definitions
                    print("Skipping atom {}: ".format(atom) + str(e))
                continue

            # TODO Add reference
            db.insert_atom_to_basisset(basset_id, element["atnum"], reference="")


def download_cgto_for_atoms(elem_list, bset_name, atnums, extra):
    """
    Obtain the contracted Gaussian functions for the basis with the
    given name, the atom with the given atomic number as well
    as the indicated extra information.

    @param elem_list   List to use for element symbol <-> atomic number lookups
    @param bset_name   Name of the basis set
    @param atnum  List of atomic numbers
    @param extra  Extra info required

    Returns a list of dicts containing the following entries:
        atnum:     atomic number
        functions: list of dict with the keys:
            angular_momentum  Angular momentum of the function
            coefficients      List of contraction coefficients
            exponents         List of contraction exponents
    """
    basis_url = json.loads(extra)["url"]

    # Lookup atomic numbers to symbols
    symbols = [elem_list[atnum]["symbol"] for atnum in atnums]

    base_url = get_base_url()
    url = base_url + "/action/portlets.BasisSetAction/template/courier_content/panel/" \
        "Main/eventSubmit_doDownload/true"

    params = {
        "bsurl": basis_url,
        "bsname": bset_name,
        "elts":  " ".join(symbols) + " ",
        "format": "Gaussian94",
        "minimize": "true",      # Get contracted version, decontraction can happen later
    }

    ret = tlsutil.get_tls_fallback(url, params=params)
    if not ret.ok:
        raise EmslError("Error getting basis set " + bset_name + " from emsl.")
    soup = BeautifulSoup(ret.text, "lxml")

    # The basis set should be encoded inside a pre tag
    if soup.pre is None:
        raise EmslError("No pre in result from emsl for basis set name " + bset_name)
    if "$bsdata" in soup.pre.text:
        raise EmslError("Only found dummy content in pre element for basis set name " +
                        bset_name)

    ret = gaussian94.loads(soup.pre.text, elem_list=elem_list)
    if len(ret) < 1:
        raise AssertionError("Something went wrong parsing EMSL basis set text "
                             "\n{}".format(soup.pre.text))
    return ret



#!/usr/bin/env python3

import shutil
import re
from .elements import iupac_list

__strip_ANSI_escapes = re.compile(r"""
  \x1b     # literal ESC
  \[       # literal [
  [;\d]*   # zero or more digits or semicolons
  [A-Za-z] # a letter
  """, re.VERBOSE).sub


def printlen(s):
    """
    Return the printed length of a string
    """
    return len(__strip_ANSI_escapes("", s))


def crop_to_printlen(s, l):
    """Return only as many characters such that the printed length
    of them is less than or equal l"""
    if printlen(s) <= l:
        return s
    i = l
    while printlen(s[:i]) < l:
        i += 1
    return s[:i]


# Flags to influence the way basis sets are listed
__display_formats_base = ["elements", "colour", "crop"]


# Make sure that both positive and negative versions exist:
available_display_formats = [f for b in __display_formats_base for f in ("no-" + b, b)]


# Ansi colour escape sequences
colours_ANSI = {
    "yellow": '\033[93m',
    "white": '\033[0m',
}


def parse_list_format(format_list):
    """Parse the format for the list_basissets function
    and return a dictionary key -> value from it"""
    ret = {}

    def negate_flag(flag):
        return flag[3:] if flag.find("no-") == 0 else "no-" + flag

    for flag in format_list:
        if flag in __display_formats_base:
            ret[flag] = True
        elif negate_flag(flag) in __display_formats_base:
            ret[negate_flag(flag)] = False
    return ret


def print_basissets(findings, highlight_atnums=[],
                    elements=False, colour=True, crop=True,
                    source_to_colour=None):
    """
    Pretty print the basissets in the list
    highlight_atnums     Highlight these elements in the list
    elements        Print the list of elements
    colour          Use colour for printing
    crop            Crop the output if it is too wide
    source_to_colour    Mapping from the source of the basis set
                        to an appropriate colour
    """
    # Colours used for display
    creset = celem = cbas = ""
    if colour:
        creset = colours_ANSI["white"]
        celem = colours_ANSI["yellow"]
        cbas = colours_ANSI["yellow"]

    # Get IUPAC element list
    elem_list = iupac_list()

    def format_element_list(basset):
        """
        Take a basis set dictionary and return a formatted string
        of the element list, taking the list of atnums to highlight into account.
        """
        atnum_symbols = [(e["atnum"], elem_list[e["atnum"]]["symbol"])
                         for e in basset["atoms"]]
        return ",".join(celem + sym + creset if atnum in highlight_atnums
                        else sym for atnum, sym in atnum_symbols)

    # Determine maximal lengths of the strings we have:
    maxlen_name = max(1, max(len(bset["name"]) for bset in findings))
    maxlen_descr = max(1, max(len(bset["description"]) for bset in findings))

    # Ignore element string length if we don't care
    if elements:
        maxlen_elem = max(printlen(format_element_list(bset)) for bset in findings)
    else:
        maxlen_elem = 0

    # Adjust depending on width of terminal
    cols, _ = shutil.get_terminal_size(fallback=(120, 50))
    cols = max(120, cols)
    extra = 4  # What we need for column separators, ...

    if maxlen_name + maxlen_descr + maxlen_elem + extra > cols:
        # We don't crop the name ever, so compute the remainder:
        rem = cols - maxlen_name - extra

        if elements:
            # 2/3 for description, but only if its needed
            # and at least 1/3 for elements:
            maxlen_descr = min(maxlen_descr, max(50, 2 * rem // 3, rem - maxlen_elem - 1))
            maxlen_elem = max(50, rem - maxlen_descr)
        else:
            maxlen_descr = rem
            maxlen_elem = 0

    # Build format string:
    fstr = cbas + "{name:" + str(maxlen_name) + "s}" + creset
    fstr += "  {description:" + str(maxlen_descr) + "s}"
    if elements:
        fstr += "  {elements:" + str(maxlen_elem) + "s}"

    for bset in findings:
        descr = bset["description"]
        elems = format_element_list(bset)
        if crop:
            if printlen(descr) > maxlen_descr:
                descr = crop_to_printlen(descr, maxlen_descr - 3)
                descr += "..."
            if printlen(elems) > maxlen_elem:
                elems = crop_to_printlen(elems, maxlen_elem - 3 + 1)
                # Remove the half-printed element number after the last ","
                elems = elems[:elems.rfind(",")]
                elems += "..."

        print(fstr.format(name=bset["name"], description=descr, elements=elems))



#!/usr/bin/env python3
from . import database as dbcache
from . import emsl, ccrepo, elements
import datetime
import os

__version__ = '0.1.1'
__all__ = ["Database", "update_database"]


class Database(dbcache.Database):
    def __init__(self, dbfile="~/.local/share/look4bas/basis_sets.db"):
        super().__init__(os.path.expanduser(dbfile))

    def lookup_basisset_full(self, basisset):
        """
        Lookup information about the basis set in the database and return
        the list of defined atoms and their basis functions.
        If the data is not stored in the database then it is automatically
        downloaded on the fly.

        @param basisset     Basis set dict as returned by search_basisset or
                            basis set id.
        """
        basisset = self.lookup_basisset(basisset)

        # TODO Check if data exists in db if not add it.

        atnums = [at["atnum"] for at in basisset["atoms"]]
        elem_list = self.lookup_element_list(basisset["source"])

        if basisset["source"] == "EMSL":
            basisset["atoms"] = emsl.download_cgto_for_atoms(elem_list, basisset["name"],
                                                             atnums, basisset["extra"])
            # TODO write data to db
        elif basisset["source"] == "ccrepo":
            basisset["atoms"] = ccrepo.download_cgto_for_atoms(elem_list,
                                                               basisset["name"],
                                                               atnums, basisset["extra"])
            # TODO write data to db
        else:
            raise ValueError("Unknown basis set source: {}".format(basisset["source"]))
        return basisset


def update_database(db):
    """
    Update the provided database, i.e. check whether new records
    exist online and update accordingly.
    """
    # TODO poor mans solution for now ... update if older than 14 days.
    #
    # Better: Check the most recent modification time for the EMSL and ccrepo
    #         data and only update the respective sources if there are changes.
    maxage = datetime.timedelta(days=14)

    age = datetime.datetime.utcnow() - db.timestamp
    if age > maxage or db.empty:
        db.clear()

        emsl.add_to_database(db)
        ccrepo.add_to_database(db)
        db.create_table_of_elements(
            "IUPAC",
            [e for e in elements.iupac_list() if e["atnum"] > 0]
        )



#!/usr/bin/env python3

from . import gaussian94
import os

""" Dictionary of the basis set formats supported by this script,
    mapped to the default file extension used.
"""
formats = {
    "Gaussian94": "g94",
}


def normalise_name(name):
    """Normalise a basis set name to yield a valid filename"""
    return name.lower().replace("/", "I").replace(" ", "_")


def save_basisset(bset, fmts, destination="."):
    """
    Save the provided basis set in the given formats
    to the provided destination directory.
    """

    for fmt in fmts:
        if fmt == "Gaussian94":
            data = gaussian94.dumps(bset["atoms"])
        else:
            raise NotImplementedError("Format {} not implemented.".format(fmt))

        path = destination + "/" + normalise_name(bset["name"]) + "." + formats[fmt]
        if os.path.exists(path):
            print("   Warn: Skipping " + path + " since file already exists")
        else:
            print("   Saving to ", path)
            with open(path, "w") as f:
                f.write(data)



#!/usr/bin/env python3

import codecs
import datetime
import os
import re
import sqlite3 as sqlite


def capitalise(word):
    return word[0].upper() + word[1:]


def quote_identifier(s, errors="strict"):
    """
    Quote identifiers for sqlite (e.g. table names)
    """
    encodable = s.encode("utf-8", errors).decode("utf-8")

    nul_index = encodable.find("\x00")

    if nul_index >= 0:
        error = UnicodeEncodeError("NUL-terminated utf-8", encodable,
                                   nul_index, nul_index + 1, "NUL not allowed")
        error_handler = codecs.lookup_error(errors)
        replacement, _ = error_handler(error)
        encodable = encodable.replace("\x00", replacement)

    return "\"" + encodable.replace("\"", "\"\"") + "\""


class Database:
    """Database frontend"""

    """Current database version"""
    database_version = 1

    def __init__(self, dbfile):
        """
        Initialise specifying database to work on
        """
        self.dbfile = dbfile
        self.conn = None

        if not os.path.isfile(dbfile):
            self.clear()
        else:
            conn = sqlite.connect(dbfile)

            # Check version: version < DB_VERSION indicates an invalid db,
            # that is too old or uninitialised and needs to be discarded
            version = conn.execute("PRAGMA user_version").fetchone()[0]
            if version is None or version < self.database_version:
                # Delete current database and create a new one
                self.clear()
            else:
                self.conn = conn
                self.__register_user_functions()

    def __register_user_functions(self):
        def matches(expr, item):
            return re.search(expr, item) is not None
        self.conn.create_function("MATCHES", 2, matches)

        def matchesi(expr, item):
            return re.search(expr, item, flags=re.I) is not None
        self.conn.create_function("MATCHESI", 2, matchesi)

    @property
    def timestamp(self):
        if os.path.exists(self.dbfile):
            return datetime.datetime.fromtimestamp(os.path.getmtime(self.dbfile))
        else:
            return datetime.datetime.fromtimestamp(0)

    def clear(self):
        """Clear the complete database and reset to untouched state"""
        if self.conn:
            self.conn.close()
            self.conn = None
        if os.path.isfile(self.dbfile):
            os.remove(self.dbfile)

        dirname = os.path.dirname(self.dbfile)
        os.makedirs(dirname, exist_ok=True)
        self.conn = sqlite.connect(self.dbfile)
        self.__register_user_functions()

        with self.conn:
            cur = self.conn.cursor()

            # Table of basis sets
            cur.execute("CREATE TABLE BasisSet("
                        "Id INTEGER PRIMARY KEY, "
                        "Name TEXT, "        # Name of the basis set
                        "Description TEXT,"  # Short description
                        "Source TEXT, "      # Source of this definition
                                             # (e.g. EMSL, ccrepo)
                        "Extra TEXT"         # Extra info, specific to the source
                        ")")

            # Table of basis definitions for atoms contained in the basis sets
            cur.execute("CREATE TABLE AtomPerBasis("
                        "Id INTEGER PRIMARY KEY, "
                        "BasisSetID INT, "   # ID of the BasSet this atom entry
                        "AtNum INT, "        # Atomic number
                        "Reference TEXT,"    # Reference of a Paper where this set of
                                             # functions for this element were defined
                        "HasFunctions INT"   # Are the basis functions stored
                                             # in the database?
                        ")")

            # Table of contracted basis functions
            cur.execute("CREATE TABLE BasisFunctions("
                        "Id INTEGER PRIMARY KEY, "
                        "AtomBasisId INT, "      # ID of the AtomBasis of this function
                        "AngularMomentum INT "   # angular momentum quantum number
                        ")")

            # Table of contraction coefficients
            cur.execute("CREATE TABLE Contraction("
                        "Id INTEGER PRIMARY KEY, "  # Unique ID
                        "FunctionId INT, "     # ID of the basis function
                        "Coefficient REAL,"    # Contraction coefficient
                        "Exponent REAL"        # Gaussian exponent
                        ")")

            # Set value to db version to indicate initialisation
            cur.execute("PRAGMA user_version = {v:d}".format(v=self.database_version))

    @property
    def empty(self):
        with self.conn:
            cur = self.conn.cursor()
            count = cur.execute("SELECT COUNT(*) FROM BasisSet").fetchone()[0]
        return count == 0

    def close(self):
        """
        Close the connection held by conn
        """
        self.conn.close()
        self.conn = None

    def create_table_of_elements(self, source, elements):
        """
        Create a table of elements in a convention used
        by a particular source (e.g. IUPAC, EMSL, ccrepo).

        elements should be a list of dicts with the entries
            atnum:   Atom number
            symbol:  Atom symbol
            name:    Atom name
        """
        tablename = quote_identifier("Elements" + str(source))
        with self.conn:
            cur = self.conn.cursor()

            # Drop the table if it exists
            cur.execute("DROP TABLE IF EXISTS " + tablename)
            cur.execute("CREATE TABLE " + tablename + " ("
                        "AtNum INTEGER PRIMARY KEY, "  # Atomic number
                        "Symbol TEXT, "  # Atom symbol in all lower case
                        "Name TEXT"      # Atom name
                        ")")

            for elem in elements:
                symbol = elem["symbol"].lower()
                cur.execute(
                    "INSERT INTO " + tablename + " "
                    "(AtNum, Symbol, Name) VALUES (?, ?, ?)",
                    (elem["atnum"], symbol, elem["name"])
                )

    def search_element(self, source, key):
        """
        Search an element by the given key in symbol,
        name and atom number and return a dict with the keys
            atnum:   Atom number
            symbol:  Atom symbol
            name:    Atom name

        @param source  The source to search in (e.g. EMSL, ccrepo, IUPAC)
        @param key     The key to search for
        """
        if isinstance(key, int):
            query = "Atnum = ?"
            args = [key]
        elif isinstance(key, str):
            query = "Name = ? OR Symbol = ?"
            args = [key.lower(), key.lower()]
        else:
            raise TypeError("Key may either be a string or an integer")

        tablename = quote_identifier("Elements" + str(source))
        with self.conn:
            cur = self.conn.cursor()

            cur.execute("SELECT name FROM sqlite_master "
                        "WHERE type='table' AND name=" + tablename)
            res = cur.fetchone()
            if res is None:
                raise ValueError("Unknown source {}".format(source))

            cur.execute("SELECT * FROM " + tablename + " " + "WHERE " + query, args)
            res = cur.fetchall()

        if len(res) == 0:
            raise ValueError("No element not found, which matches key {}".format(key))
        assert len(res) == 1
        entry = res[0]

        symbol = capitalise(entry[1])
        return {"atnum": entry[0], "symbol": symbol, "name": entry[2]}

    def lookup_element_list(self, source):
        """
        Build the list of elements per atomic number for a particular
        source (e.g. EMSL, ccrepo, IUPAC).

        The first entry of the list is "X", which is a dummy place holder.
        """
        tablename = quote_identifier("Elements" + str(source))
        with self.conn:
            cur = self.conn.cursor()

            cur.execute("SELECT name FROM sqlite_master "
                        "WHERE type='table' AND name=" + tablename)
            res = cur.fetchone()
            if res is None:
                raise ValueError("Unknown source {}".format(source))

            cur.execute("SELECT AtNum, Symbol, Name FROM " + tablename +
                        " ORDER BY AtNum ASC")
            ret = [{"atnum": 0, "symbol": "X", "name": "dummy"}]
            for atnum, symbol, name in cur.fetchall():
                ret.append({"atnum": atnum, "name": name,
                            "symbol": capitalise(symbol)})
        return ret

    def insert_basis_function(self, atbas_id, angular_momentum, coefficients, exponents):
        """
        Insert a list of contractions for the provided element.

        @param atbas_id      ID of the atom for which functions should be added to the
                             basis set
        @param angular_momentum    Angular momentum of the contracted basis function
                                   to be inserted
        @param coefficients  List of contraction coefficients
        @param exponents     List of contraction exponents
        """
        if not isinstance(atbas_id, int):
            raise TypeError("atbas_id needs to be an integer")
        if not isinstance(angular_momentum, int):
            raise TypeError("angular_momentum needs to be an integer")
        if len(coefficients) != len(exponents):
            raise ValueError("Coefficients and exponents need to have the "
                             "same length")
        with self.conn:
            cur = self.conn.cursor()

            cur.execute(
                "INSERT INTO BasisFunctions (AtomBasisId, AngularMomentum)"
                "VALUES (?, ?)", (atbas_id, angular_momentum)
            )
            function_id = cur.execute("SELECT last_insert_rowid()").fetchone()[0]

            for i, coeff in enumerate(coefficients):
                exp = exponents[i]
                cur.execute(
                    "INSERT INTO Contraction "
                    "(FunctionId, Coefficient, Exponent) VALUES"
                    "(?, ?, ?)", (function_id, coeff, exp)
                )

            # Mark that the appropriate element has basis functions set in the db
            cur.execute("UPDATE AtomPerBasis SET HasFunctions = 1 WHERE Id = ?", atbas_id)

    def lookup_basis_functions(self, atbas_id):
        """
        Select all basis functions belonging to a particular atbas_id.

        Returns a list of dicts with the keys
            angular_momentum  Angular momentum of the function
            coefficients      List of contraction coefficients
            exponents         List of contraction exponents
        """
        if not isinstance(atbas_id, int):
            raise TypeError("atbas_id needs to be an integer")

        with self.conn:
            cur = self.conn.cursor()
            cur.execute("SELECT BasisFunctions.AngularMomentum, Contraction.FunctionId, "
                        "Contraction.Coefficient, Contraction.Exponent "
                        "FROM BasisFunctions "
                        "INNER JOIN Contraction ON BasisFunctions.Id = "
                        "Contraction.FunctionId WHERE BasisFunctions.AtomBasisId = ?",
                        (str(atbas_id),))
            contractions = cur.fetchall()

            ret = {}
            for am, fun_id, coeff, exp in contractions:
                if fun_id not in ret:
                    ret[fun_id] = {"coefficients": [coeff],
                                   "exponents": [exp],
                                   "angular_momentum": am, }
                else:
                    assert am == ret[fun_id]["angular_momentum"]
                    ret[fun_id]["coefficients"].append(coeff)
                    ret[fun_id]["exponents"].append(exp)
        return list(ret.values())

    def insert_atom_to_basisset(self, basset_id, atnum, reference=""):
        """
        Insert a new atom for a particular basis set.

        @param basset_id   ID of the basis set
        @param atnum       Atomic number
        @param reference   A paper reference if available

        returns the id of the element which was inserted.
        """
        if not isinstance(basset_id, int):
            raise TypeError("basset_id needs to be an integer")
        if not isinstance(atnum, int):
            raise TypeError("atnum needs to be an integer")
        if not isinstance(reference, str):
            raise TypeError("reference needs to be a string")

        with self.conn:
            cur = self.conn.cursor()
            cur.execute(
                "INSERT INTO AtomPerBasis (BasisSetID, AtNum, Reference, HasFunctions)"
                "VALUES (?, ?, ?, 0)", (basset_id, atnum, reference)
            )
            rowid = cur.execute("SELECT last_insert_rowid()").fetchone()[0]
        return rowid

    def insert_basisset(self, name, source, extra="", description=""):
        """
        Insert a new basis set.

        @param name        Name of the basis set
        @param source      Source where the data comes from (e.g. EMSL, ccrepo)
        @param extra       Extra data depending on the source
        @param description Description of the basis set
        """
        if not isinstance(name, str):
            raise TypeError("name needs to be a string")
        if not isinstance(description, str):
            raise TypeError("description needs to be a string")
        if not isinstance(source, str):
            raise TypeError("source needs to be a string")
        if not isinstance(extra, str):
            raise TypeError("extra needs to be a string")

        with self.conn:
            cur = self.conn.cursor()
            cur.execute(
                "INSERT INTO BasisSet (Name, Description, Source, Extra)"
                "VALUES (?, ?, ?, ?)", (name, description, source, extra)
            )
            rowid = cur.execute("SELECT last_insert_rowid()").fetchone()[0]
        return rowid

    def __ditcify_basisset_query_result(self, res):
        ret = {}
        for row in res:
            basset = ret.get(row[0], {"atoms": []})
            basset["id"], basset["name"], basset["description"], \
                basset["source"], basset["extra"], \
                atbas_id, atnum, has_functions = row

            if basset["id"] is None or atnum is None:
                continue  # Skip rows with undefined fields

            basset["atoms"].append({
                "atnum": atnum,
                "atbas_id": atbas_id,
                "has_functions": bool(has_functions)
            })
            ret[row[0]] = basset
        return list(ret.values())

    def lookup_basisset(self, basisset):
        """
        Lookup information about the basis set in the database and return
        the list of defined atoms and their atbas_id to perform further queries
        with lookup_basis_functions. If the has_functions flag is false,
        then there are no basis functions defined in the database for this
        combination of atom and basis set. Use lookup_basisset_full
        to amend the database by downloading information from the appropriate
        internet source.

        @param basisset     Basis set dict as returned by search_basisset or
                             basis set id.
        """
        if isinstance(basisset, int):
            basset_id = basisset
        elif isinstance(basisset, dict):
            if "id" not in basisset:
                raise ValueError("Passed dict is not a basis set dict as returned "
                                 "by lookup_basis_functions")
            basset_id = basisset["id"]
        else:
            raise TypeError("basset_id needs to be an integer or a dict")

        with self.conn:
            cur = self.conn.cursor()

            cur.execute("SELECT BasisSet.Id, BasisSet.Name, BasisSet.Description, " +
                        "BasisSet.Source, BasisSet.Extra, AtomPerBasis.Id, " +
                        "AtomPerBasis.AtNum, AtomPerBasis.HasFunctions " +
                        "FROM BasisSet LEFT JOIN AtomPerBasis " +
                        "ON AtomPerBasis.BasisSetID = BasisSet.Id " +
                        "WHERE BasisSet.Id = ?", (str(basset_id),))
            ret = self.__ditcify_basisset_query_result(cur.fetchall())
            assert len(ret) == 1
            return ret[0]

    def lookup_basisset_full(self, basset_id):
        raise NotImplementedError("Not implemented in the basic database. "
                                  "Use api.Database object for this purpose.")

    def search_basisset(self, name=None, description=None, ignore_case=False,
                        has_atnums=[], source=None, regex=False, pattern=None):
        """
        Function to filter basis sets. If no arguments are provided,
        all registered basis sets will be returned.

        name    String to be contained in the basis set same
                or regular expression to be matched against the name.
        description   String to be contained in the description
                      or regular expression to be matched against it.
        pattern      String to be contained either in the basis set name
                     *or* the description or regex to be matching
                     against either these fields.
        has_atnums   Atoms to be contained in this basis set
                     Should be a list of atomic numbers.
        source       The source of the basis set. Is matched exactly.
        ignore_case  Regular expression and string matchings
                     in name, pattern and description are done ignoring case.
        regex        Are the strings supplied to name, descriptions
                     and pattern
                     to be interpreted as regular exrpessions

        Returns a list of dicts with content id, name, description,
        source, extra and their respective atoms.
        """
        if name is not None and not isinstance(name, str):
            raise TypeError("name needs to be None or a string")
        if description is not None and not isinstance(description, str):
            raise TypeError("descrption needs to be None or a string")
        if pattern is not None and not isinstance(pattern, str):
            raise TypeError("pattern needs to be None or a string")
        if source is not None and not isinstance(source, str):
            raise TypeError("source needs to be None or a string")
        if not isinstance(has_atnums, list):
            raise TypeError("has_atnums needs to be alist")

        if regex:
            if ignore_case:
                def match_field(field):
                    return "matchesi(?, " + field + ")"
            else:
                def match_field(field):
                    return "matches(?, " + field + ")"
        else:
            if ignore_case:
                def match_field(field):
                    return "instr(lower(" + field + "), lower(?))"
            else:
                def match_field(field):
                    return "instr(" + field + ", ?)"

        prefix = ("SELECT BasisSet.Id, BasisSet.Name, BasisSet.Description, " +
                  "BasisSet.Source, BasisSet.Extra, AtomPerBasis.Id, " +
                  "AtomPerBasis.AtNum, AtomPerBasis.HasFunctions " +
                  "FROM BasisSet LEFT JOIN AtomPerBasis " +
                  "ON AtomPerBasis.BasisSetID = BasisSet.Id ")
        wheres = []
        args = []

        if name is not None:
            wheres.append(match_field("Name"))
            args.append(name)
        if description:
            wheres.append(match_field("Description"))
            args.append(description)
        if pattern:
            q = "( " + match_field("Description") + \
                " OR " + match_field("Name") + " )"
            wheres.append(q)
            args.append(pattern)
            args.append(pattern)
        if source:
            wheres.append("Source = ?")
            args.append(source)
        if has_atnums:
            for atnum in has_atnums:
                wheres.append(
                    "BasisSet.Id IN (SELECT BasisSetID FROM AtomPerBasis WHERE AtNum = ?)"
                )
                if not isinstance(atnum, int):
                    raise TypeError("All entries of has_atnums need to be integers")
                args.append(atnum)

        if wheres:
            query = prefix + " WHERE " + " AND ".join(wheres)
        else:
            query = prefix

        with self.conn:
            cur = self.conn.cursor()
            cur.execute(query, args)
            return self.__ditcify_basisset_query_result(cur.fetchall())



#!/usr/bin/env python3
from . import elements


NUMBER_TO_AM = ["S", "P", "D", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O"]


def __float_fortran(string):
    return float(string.replace("d", "e").replace("D", "E"))


def __strip_comments(string):
    """
    Strip all comments from the string
    """
    return "\n".join(e.strip() for e in string.split("\n") if e and e[0] != '!')


def __parse_contractions(lines, *outputs):
    for line in lines:
        exp_coeffs = line.split()
        if len(exp_coeffs) != len(outputs):
            raise ValueError("Not enough columns found. Expected {} columns, "
                             "but found only {}. Culprit line is "
                             "'{}'".format(len(outputs), len(exp_coeffs), line))
        try:
            for i in range(len(outputs)):
                outputs[i].append(__float_fortran(exp_coeffs[i]))
        except ValueError:
            raise ValueError("Could not convert columns to float. "
                             "Culprit line is '{}'".format(line))


def __parse_element_block(block, elem_symbols_lower):
    ret = {"functions": []}
    lines = [__strip_comments(l) for l in block.split("\n")]
    lines = [l for l in lines if len(l) > 0]

    symbol, _ = lines[0].split(maxsplit=1)
    try:
        ret["atnum"] = elem_symbols_lower.index(symbol.lower())
    except KeyError:
        raise ValueError("Element block starting with invalid element symbol "
                         "{}".format(symbol))

    idx = 1
    while idx < len(lines):
        # New function definition starts with angular momentum
        # and the number of contractions
        amstr, n_contr, _ = lines[idx].split(maxsplit=2)
        amstr = amstr.upper()

        try:
            n_contr = int(n_contr)
        except ValueError:
            raise ValueError("Expect number of contracted function (following AM letter) "
                             "to be an integer, not {}".format(n_contr))

        if amstr == "sp":
            # This is a special case, where an s and p shell are defined
            # at the same time.

            s_coefficients = []
            p_coefficients = []
            exponents = []
            __parse_contractions(lines[idx + 1:idx + n_contr + 1],
                                 exponents, s_coefficients, p_coefficients)

            ret["functions"].append({
                "angular_momentum": 0,
                "coefficients": s_coefficients,
                "exponents": exponents,
            })
            ret["functions"].append({
                "angular_momentum": 1,
                "coefficients": p_coefficients,
                "exponents": exponents,
            })
            idx += n_contr + 1
        else:
            # Standard cases:
            try:
                am = NUMBER_TO_AM.index(amstr)
            except ValueError:
                raise ValueError("Invalid angular momentum string {}".format(amstr))

            coefficients = []
            exponents = []
            __parse_contractions(lines[idx + 1:idx + n_contr + 1], exponents,
                                 coefficients)

            ret["functions"].append({
                "angular_momentum": am,
                "coefficients": coefficients,
                "exponents": exponents,
            })
            idx += n_contr + 1
    return ret


def loads(string, elem_list=elements.iupac_list()):
    """
    Parse a string, which represents a basis set file in g94
    format and return the defined basis functions
    as a a list of dicts containing the following entries:
        atnum:     atomic number
        functions: list of dict with the keys:
            angular_momentum  Angular momentum of the function
            coefficients      List of contraction coefficients
            exponents         List of contraction exponents

    The format is roughly speaking:
      ****
      Element_symbol
      AM   n_contr
         exp1     coeff1
         exp2     coeff2
      AM2  n_contr
         exp1     coeff1
         exp2     coeff2
      ****
      Element_symbol
      ...

    where
       Element_symbol    a well-known element symbol
       AM                azimuthal quantum number as a symbol for angular momentum
       n_contr           number of contracted primitives
       exp1              exponents
       coeff1            corresponding contraction coefficients.

    All element blocks are separated by '****'. The numbers for exp and coeff
    may be given in the Fortran float convention (using 'D's instead of 'E's).
    """

    # First split into blocks at the separating "****)
    blocks = [b.strip() for b in string.split("****")]

    if len(blocks) < 2:
        raise ValueError("At least one '****' sequence in the input string is expected")
    if len(__strip_comments(blocks[0])) > 0:
        raise ValueError("Found valid content before initial '****' sequence")
    if len(__strip_comments(blocks[-1])) > 0:
        raise ValueError("Found valid content after final '****' sequence")

    # Convert the elem_list to all-lower-case
    elem_symbols_lower = [e["symbol"].lower() for e in elem_list]

    # The first and last block are just comments or trailing newlines and can
    # be ignored
    return [__parse_element_block(block, elem_symbols_lower) for block in blocks[1:-1]]


def dumps(data, elem_list=elements.iupac_list()):
    """
    Take a list of dicts containing the entries
        atnum:     atomic number
        functions: list of dict with the keys:
            angular_momentum  Angular momentum of the function
            coefficients      List of contraction coefficients
            exponents         List of contraction exponents
    and dump a string representing this basis set definition
    in Gaussian94 format.
    """
    lines = []
    for atom in data:
        lines.append("****")
        lines.append(elem_list[atom["atnum"]]["symbol"] + "     0")
        for fun in atom["functions"]:
            lfun = len(fun["coefficients"])
            if lfun != len(fun["exponents"]):
                raise ValueError("Length of coefficients and length of exponents "
                                 "in contraction specification need to agree")

            am = NUMBER_TO_AM[fun["angular_momentum"]]
            lines.append(am + "   " + str(lfun) + "   1.00")

            for i, coeff in enumerate(fun["coefficients"]):
                exp = fun["exponents"][i]
                fmt = "{0:15.7f}             {1: #11.8G}"
                lines.append(fmt.format(exp, coeff))
    lines.append("****\n")
    return "\n".join(lines)



#!/usr/bin/env python3

dbfile = "~/.local/share/look4bas/basis_sets.db"
default_download_formats = ["Gaussian94"]
list_formats = {
    "default": ["crop", "noelements", "colour"],
    "extra":   ["crop", "elements", "colour"]
}




import requests
import requests.adapters
import ssl


class TLSLowerAdapter(requests.adapters.HTTPAdapter):
    def init_poolmanager(self, *args, **kwargs):
        kwargs['ssl_context'] = ssl.SSLContext(self.ssl_version)
        return super().init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        kwargs['ssl_context'] = ssl.SSLContext(self.ssl_version)
        return super().proxy_manager_for(*args, **kwargs)


class TLSv1Adapter(TLSLowerAdapter):
    ssl_version = ssl.PROTOCOL_TLSv1


class TLSv1_1Adapter(TLSLowerAdapter):
    ssl_version = ssl.PROTOCOL_TLSv1_1


class TLSv1_2Adapter(TLSLowerAdapter):
    ssl_version = ssl.PROTOCOL_TLSv1_2


def method_tls_fallback(url, method, *args, **kwargs):
    """
    Try to perform a method on an url using requests.
    If the get request fails due to an SSLError, we try to lower
    the TLS version until it finally succeeds.
    """
    # Try to get as-is
    try:
        session = requests.Session()
        return getattr(session, method)(url, *args, **kwargs)
    except requests.exceptions.SSLError:
        pass

    err = None
    for adapter in [TLSv1_2Adapter, TLSv1_1Adapter, TLSv1Adapter]:
        try:
            session = requests.Session()
            session.mount("https://", adapter())
            return getattr(session, method)(url, *args, **kwargs)
        except requests.exceptions.SSLError as e:
            err = e
    raise err


def get_tls_fallback(url, *args, **kwargs):
    return method_tls_fallback(url, "get", *args, **kwargs)


def post_tls_fallback(url, *args, **kwargs):
    return method_tls_fallback(url, "post", *args, **kwargs)



"""Setup for look4bas"""

# Use setuptools for these commands (they don't work well or at all
# with distutils).  For normal builds use distutils.
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='look4bas',
    description="Search for Gaussian basis sets",
    long_description='Command line tool to search for contracted Gaussian-type basis '
    'sets in electronic structure theory',
    #
    url='https://github.com/mfherbst/look4bas',
    author='Michael F. Herbst',
    author_email="info@michael-herbst.com",
    license="GPL 3",
    #
    packages=['look4bas'],
    scripts=["bin/look4bas"],
    version='0.1.1',
    #
    python_requires='>=3',
    install_requires=[
        'requests (>=2.2)',
        'beautifulsoup4 (>= 4.2)',
        'lxml (>= 4.2)'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Intended Audience :: Science/Research',
        "Topic :: Scientific/Engineering :: Chemistry",
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: Unix',
    ],
)



#!/usr/bin/env python3
import collections


def iupac_list():
    """
    Build the list of elements per atomic number according to the IUPAC
    convention.

    The first entry of the list is "X", which is a dummy place holder.
    """
    Element = collections.namedtuple("Element", ["atnum", "symbol", "name"])

    elist = [
      Element(  0, "X",  "dummy"),
      Element(  1, "H" , "hydrogen"),       Element(  2, "He", "helium"),
      Element(  3, "Li", "lithium"),        Element(  4, "Be", "beryllium"),
      Element(  5, "B" , "boron"),          Element(  6, "C",  "carbon"),
      Element(  7, "N" , "nitrogen"),       Element(  8, "O",  "oxygen"),
      Element(  9, "F" , "fluorine"),       Element( 10, "Ne", "neon"),
      Element( 11, "Na", "sodium"),         Element( 12, "Mg", "magnesium"),
      Element( 13, "Al", "aluminium"),      Element( 14, "Si", "silicon"),
      Element( 15, "P" , "phosphorus"),     Element( 16, "S",  "sulphur"),
      Element( 17, "Cl", "chlorine"),       Element( 18, "Ar", "argon"),
      Element( 19, "K" , "potassium"),      Element( 20, "Ca", "calcium"),
      Element( 21, "Sc", "scandium"),       Element( 22, "Ti", "titanium"),
      Element( 23, "V" , "vanadium"),       Element( 24, "Cr", "chromium"),
      Element( 25, "Mn", "manganese"),      Element( 26, "Fe", "iron"),
      Element( 27, "Co", "cobalt"),         Element( 28, "Ni", "nickel"),
      Element( 29, "Cu", "copper"),         Element( 30, "Zn", "zinc"),
      Element( 31, "Ga", "gallium"),        Element( 32, "Ge", "germanium"),
      Element( 33, "As", "arsenic"),        Element( 34, "Se", "selenium"),
      Element( 35, "Br", "bromine"),        Element( 36, "Kr", "krypton"),
      Element( 37, "Rb", "rubidium"),       Element( 38, "Sr", "strontium"),
      Element( 39, "Y" , "yttrium"),        Element( 40, "Zr", "zirconium"),
      Element( 41, "Nb", "niobium"),        Element( 42, "Mo", "molybdenum"),
      Element( 43, "Tc", "technetium"),     Element( 44, "Ru", "ruthenium"),
      Element( 45, "Rh", "rhodium"),        Element( 46, "Pd", "palladium"),
      Element( 47, "Ag", "silver"),         Element( 48, "Cd", "cadmium"),
      Element( 49, "In", "indium"),         Element( 50, "Sn", "tin"),
      Element( 51, "Sb", "antimony"),       Element( 52, "Te", "tellurium"),
      Element( 53, "I" , "iodine"),         Element( 54, "Xe", "xenon"),
      Element( 55, "Cs", "caesium"),        Element( 56, "Ba", "barium"),
      Element( 57, "La", "lanthanum"),      Element( 58, "Ce", "cerium"),
      Element( 59, "Pr", "praseodymium"),   Element( 60, "Nd", "neodymium"),
      Element( 61, "Pm", "promethium"),     Element( 62, "Sm", "samarium"),
      Element( 63, "Eu", "europium"),       Element( 64, "Gd", "gadolinium"),
      Element( 65, "Tb", "terbium"),        Element( 66, "Dy", "dysprosium"),
      Element( 67, "Ho", "holmium"),        Element( 68, "Er", "erbium"),
      Element( 69, "Tm", "thulium"),        Element( 70, "Yb", "ytterbium"),
      Element( 71, "Lu", "lutetium"),       Element( 72, "Hf", "hafnium"),
      Element( 73, "Ta", "tantalum"),       Element( 74, "W", "tungsten"),
      Element( 75, "Re", "rhenium"),        Element( 76, "Os", "osmium"),
      Element( 77, "Ir", "iridium"),        Element( 78, "Pt", "platinum"),
      Element( 79, "Au", "gold"),           Element( 80, "Hg", "mercury"),
      Element( 81, "Tl", "thallium"),       Element( 82, "Pb", "lead"),
      Element( 83, "Bi", "bismuth"),        Element( 84, "Po", "polonium"),
      Element( 85, "At", "astatine"),       Element( 86, "Rn", "radon"),
      Element( 87, "Fr", "francium"),       Element( 88, "Ra", "radium"),
      Element( 89, "Ac", "actinium"),       Element( 90, "Th", "thorium"),
      Element( 91, "Pa", "protactinium"),   Element( 92, "U", "uranium"),
      Element( 93, "Np", "neptunium"),      Element( 94, "Pu", "plutonium"),
      Element( 95, "Am", "americium"),      Element( 96, "Cm", "curium"),
      Element( 97, "Bk", "berkelium"),      Element( 98, "Cf", "californium"),
      Element( 99, "Es", "einsteinium"),    Element(100, "Fm", "fermium"),
      Element(101, "Md", "mendelevium"),    Element(102, "No", "nobelium"),
      Element(103, "Lr", "lawrencium"),     Element(104, "Rf", "rutherfordium"),
      Element(105, "Db", "dubnium"),        Element(106, "Sg", "seaborgium"),
      Element(107, "Bh", "bohrium"),        Element(108, "Hs", "hassium"),
      Element(109, "Mt", "meitnerium"),     Element(110, "Ds", "darmstadtium"),
      Element(111, "Rg", "roentgenium"),    Element(112, "Cn", "copernicium"),
      Element(113, "Nh", "nihonium"),       Element(114, "Fl", "flerovium"),
      Element(115, "Mc", "moscovium"),      Element(116, "Lv", "livermorium"),
      Element(117, "Ts", "tennessine"),     Element(118, "Og", "oganesson"),
    ]
    return [dict(e._asdict()) for e in elist]



#!/usr/bin/env python3

from bs4 import BeautifulSoup
from . import tlsutil, gaussian94
import json
import re

"""ccrepo base url"""
base_url = "https://grant-hill.group.shef.ac.uk/ccrepo"
# https does not work


class CcrepoError(Exception):
    """
    Generic error thrown if some data obtained from the EMSL basis set exchange
    is not in the format expected
    """
    def __init__(self, message):
        super(CcrepoError, self).__init__(message)


def get_element_list():
    ret = tlsutil.get_tls_fallback(base_url)
    if not ret.ok:
        raise CcrepoError("Error downloading list of elements from ccrepo")
    soup = BeautifulSoup(ret.text, "lxml")

    table = soup.find_all(id="pertable")
    if len(table) != 1:
        raise CcrepoError("Found more than one periodic table on the page")
    table = table[0]

    elements = []
    for elem in table.find_all(class_=re.compile("xs|xp|xd|xf|xg")):
        atnum = elem.find(class_="at_num")
        sym = elem.find(class_="symbol")

        if atnum is None or sym is None:
            continue

        try:
            atnum = int(atnum.text)
        except TypeError as e:
            raise CcrepoError("Cannot interpret as atom number: " + str(e))

        if "href" not in sym.a.attrs:
            raise CcrepoError("No element link fund for " + sym.a.text)
        name = sym.a["href"]
        if name.endswith("index.html"):
            name = name[:-10]
        name = name.strip("/")

        sym = sym.text
        elem_obj = {"symbol": sym, "name": name, "atnum": atnum}
        elements.append(elem_obj)
    return elements


def get_basis_g94(element, basis):
    """
    Get info about a basis set for a particular
    element in Gaussian94 format

    element: Element dictionary
    basis:   basis set identifier as used internally
             by ccrepo (returned as the mapped string by
             get_basis_sets_for_element.

    returns a dict with
        reference     Reference to the basis set
        description   Basis set description
        definiton     The basis set definition in Gaussian94 format
    """
    sym = element["symbol"].lower()
    ele = element["name"]

    payload = {"basis": basis, "program": "Gaussian"}
    page = base_url + "/" + ele + "/" + sym + "basis.php"
    ret = tlsutil.post_tls_fallback(page, data=payload)

    if not ret.ok:
        raise CcrepoError("Error downloading page " + page)

    if len(ret.text) == 0:
        raise CcrepoError("Got unexpected empty page on " + page)

    soup = BeautifulSoup(ret.text, "lxml")
    cont = soup.find_all(class_="container")
    if len(cont) == 0:
        raise CcrepoError("Found no container on page " + page)
    if len(cont) > 1:
        raise CcrepoError("Found more than one container on the page " + page)
    cont = cont[0]

    # All basis set definition content sits in a nobr block
    nobr = str(cont.nobr)
    nobr = nobr.replace("<nobr>", "")
    nobr = nobr.replace("</nobr>", "")
    nobr = nobr.replace("<br/>", "\n")
    nobr = nobr.replace("\n\n", "\n")
    nobr = re.sub("[ \t\r\f\v\xa0]", " ", nobr)
    nobr = nobr.replace("\n ", "\n")
    nobr = nobr.strip("\n")

    # Replace the BASIS= line by ****
    definition = re.sub("\nBASIS=[^\n]+\n", "\n****\n", nobr)

    # Find prelines with reference and description
    # This is really messy, but essentially tries to
    # extract the first two lines of real text
    cont_text = str(cont)
    cont_text = cont_text.replace("<br/>", "\n")
    prelines = []
    for line in cont_text.split("\n"):
        if len(line) == 0 or 'class="container"' in line:
            continue
        prelines.append(line.strip())
        if len(prelines) == 2:
            break

    reference = prelines[0]
    description = prelines[1]

    # Post-processing: Remove 'for Element':
    ifor = description.rfind("for ")
    description = description[:ifor].strip()
    description = description.replace("  ", " ")

    return {
        "reference": reference,
        "description": description,
        "definition": definition,
    }


def get_basis_sets_for_elem(element):
    page = base_url + "/" + element["name"] + "/index.html"
    ret = tlsutil.get_tls_fallback(page)
    if not ret.ok:
        raise CcrepoError("Error downloading list of elements from: " +
                          element["name"] + "/index.html")
    soup = BeautifulSoup(ret.text, "lxml")

    opt = soup.find_all(id="basis")
    if len(opt) == 0:
        pagetext = soup.text.strip()
        if "not quite ready to go yet" in pagetext or \
           "no correlation consistent basis sets" in pagetext:
            # The page is not yet ready ... return empty dictionary
            return dict()
        else:
            raise CcrepoError("Could not find basis on page.")
    elif len(opt) > 1:
        raise CcrepoError("Found more than one basis "
                          " field on the page " + page)
    opt = opt[0]
    return {option.text: option["value"] for option in opt.find_all("option")}


def add_to_database(db):
    """
    Add the basis set definitions to the database
    """
    elements = get_element_list()
    db.create_table_of_elements("ccrepo", elements)

    # Obtain unique list of basis sets and the elements
    # these are defined for
    bases = dict()
    for elem in elements:
        bas = get_basis_sets_for_elem(elem)

        for name in bas:
            if name in bases:
                bases[name]["atoms"].append(elem["atnum"])
            else:
                # Download the basis for this very element
                # to obtain the description string
                basdef = get_basis_g94(elem, bas[name])

                bases[name] = {
                    "name": name,
                    "key":  bas[name],
                    "atoms": [elem["atnum"]],
                    "description": basdef["description"],
                }
    bases = list(bases.values())

    # Now add all of these to the database:
    for basset in bases:
        extra = json.dumps({"key": basset["key"]})
        basset_id = db.insert_basisset(basset["name"],
                                       description=basset["description"],
                                       source="ccrepo", extra=extra)
        for atnum in basset["atoms"]:
            # TODO Add reference
            db.insert_atom_to_basisset(basset_id, atnum, reference="")


def download_cgto_for_atoms(elem_list, bset_name, atnums, extra):
    """
    Obtain the contracted Gaussian functions for the basis with the
    given name, the atom with the given atomic number as well
    as the indicated extra information.

    @param elem_list   List to use for element symbol <-> atomic number lookups
    @param bset_name   Name of the basis set
    @param atnum  List of atomic numbers
    @param extra  Extra info required

    Returns a list of dicts containing the following entries:
        atnum:     atomic number
        functions: list of dict with the keys:
            angular_momentum  Angular momentum of the function
            coefficients      List of contraction coefficients
            exponents         List of contraction exponents
    """
    key = json.loads(extra)["key"]

    ret = []
    for atnum in atnums:
        basdef = get_basis_g94(elem_list[atnum], key)

        # Parse obtained data and append to ret
        basparsed = gaussian94.loads(basdef)
        assert len(basparsed) == 1
        ret.append(basparsed[0])
    return ret



#!/usr/bin/env python3


import argparse
from look4bas import display, store, config, elements
import look4bas


def add_cmd_args_to(parser):
    parser.add_argument("--dbfile", default=config.dbfile,
                        help="File where the database of basis set "
                        "information is cached.")
    parser.add_argument("--destination", default=".", type=str, metavar="directory",
                        help="When downloading basis sets using --download store them in "
                        "this directory. (Default: '.', i.e. the current working "
                        "directory")

    mode = parser.add_mutually_exclusive_group()
    mode.add_argument("--list", action='store_true', help="List the matching basis "
                      "sets (Default)")
    mode.add_argument("--download", nargs="*", metavar='format',
                      choices=store.formats,
                      help="Download the matching basis sets in the requested formats "
                      "(Default: " + " ".join(config.default_download_formats) + ") " +
                      "to the directory specified by --destination. ")

    #
    # Filter options
    #
    filters = parser.add_argument_group("Basisset filters")
    filters.add_argument("pattern", nargs='?', default=None, type=str,
                         help="A regular expression to match against the basis set name "
                         "*or* the description. So if either description or basis set "
                         "name matches this, the basis set is considered a match.")
    filters.add_argument("--elements", metavar="element", nargs='+',
                         help="List of elements the basis set should contain. "
                         "Implies '--format elements'.")
    filters.add_argument("--regexp", "-e", dest="name_regexp", metavar="regexp",
                         help="A regular expression to match *only* against the basis "
                         "set name.")
    filters.add_argument("--description-regexp", "-d", metavar="regexp",
                         dest="description_regexp",
                         help="Regular expression the basis set description "
                         "should match.")
    filters.add_argument("--ignore-case", "-i", action="store_true", dest="ignorecase",
                         help="Ignore case when matching patterns")

    #
    # Formatting options
    #
    list_formats = parser.add_argument_group("Formatting options")
    list_formats.add_argument("--extra", action="store_true",
                              help="Use the 'extra' format style when listing basis "
                              "sets. This is currently defined as '" +
                              " ".join(config.list_formats["extra"]) + "'")
    list_formats.add_argument("--format",
                              choices=display.available_display_formats, nargs="+",
                              metavar="flag",
                              help="Amend the default list formatting by specifying "
                              "further flags.")
    return parser


def cmd_post_parsing_cleanup(args):
    # TODO Ideally this should be directly integrated into the
    #      parsing done by argparse ... but I cannot be bothered right now.

    # If filtering for elements is requested, transform elements
    # to atomic numbers
    elem_list = elements.iupac_list()
    elem_symbols_lower = [e["symbol"].lower() for e in elem_list]

    def to_atnum(sym):
        return elem_symbols_lower.index(sym.lower())

    if args.elements:
        args.elements = set(
            to_atnum(e) for spec in args.elements for e in spec.split(",")
            if len(e) > 0
        )


def lookup_basissets(db, args):
    kwargs = {"regex": True, "ignore_case": bool(args.ignorecase), }

    # Build filters
    if args.elements:
        kwargs["has_atnums"] = list(args.elements)
    if args.name_regexp:
        kwargs["name"] = args.name_regexp
    if args.description_regexp:
        kwargs["description"] = args.description_regexp
    if args.pattern:
        # Matches if name *or* description is matched
        kwargs["pattern"] = args.pattern
    return db.search_basisset(**kwargs)


def display_results(args, findings):
    # display_args are the kwargs for the display function
    # which will be called further below.
    if args.elements:
        display_args = {
            "elements": True,
            "highlight_atnums": list(args.elements),
        }
    else:
        display_args = {}

    # Build format list, by considering default and extra
    # and extending further flags if the user wants this
    list_format = config.list_formats["default"]
    if args.extra:
        list_format = config.list_formats["extra"]
    if args.format is not None:
        list_format.extend(args.format)
    display_args.update(display.parse_list_format(list_format))

    display.print_basissets(findings, **display_args)


def download_results(args, db, findings):
    # Append at least the default format to download
    args.download.extend(config.default_download_formats)

    # TODO Later we probably want a more elaborate selection
    #      mechanism where one can select amongst the matches
    #      for those to be downloaded
    #
    #      I.e. ideally we would have some interactive selection
    #      mechanism here, where maybe file names can be changed
    #
    #      or basis sets can be decontracted or ...

    for bset in findings:
        # TODO One could maybe use colour here as well with
        #      the colour scheme here and on display matching up
        print("Obtaining {:50s} (from {})".format(bset["name"], bset["source"]))
        bset = db.lookup_basisset_full(bset)
        store.save_basisset(bset, args.download, args.destination)


def main():
    # Parse the commandline arguments
    parser = argparse.ArgumentParser(
        description="Commandline tool to search and download Gaussian basis sets. "
        "The tool downloads (and caches) the list of basis sets from the emsl basis set "
        "exchange (https://bse.pnl.gov/bse/portal) for offline search and allows to "
        "easily download individual basis sets from it.",
    )
    add_cmd_args_to(parser)
    args = parser.parse_args()
    cmd_post_parsing_cleanup(args)

    # Setup database:
    db = look4bas.Database(args.dbfile)
    look4bas.update_database(db)

    # Search for basis sets
    findings = lookup_basissets(db, args)
    if not findings:
        raise SystemExit("No basis set matched your search")

    # Since args.download may be absent or present, but without a value,
    # we cannot plainly use 'if args.download' here.
    if args.download is not None:
        download_results(args, db, findings)
    else:
        display_results(args, findings)


if __name__ == "__main__":
    main()
